VERSION='1.13.0'

from cx_Freeze import setup, Executable
import os
import sys
import distutils
import shutil
import glob

include_files = [
    ('data/pdfarranger.ui', 'share/pdfarranger/pdfarranger.ui'),
    ('data/menu.ui', 'share/pdfarranger/menu.ui'),
    ('data/icons/hicolor/scalable', 'share/icons/hicolor/scalable'),
    ('build/mo', 'share/locale'),
]


def clean_build():
    dirs = os.listdir('build')
    keep = ['mo', 'lib']
    for d in dirs:
        if d not in keep:
            shutil.rmtree(os.path.join('build', d))


clean_build()


def addfile(relpath, warn_missing=False):
    f = os.path.join(sys.prefix, relpath)
    if warn_missing and not os.path.isfile(f):
        print("{} cannot be found.".format(f), file=sys.stderr)
    else:
        include_files.append((f, relpath))


def addlocale(name):
    langs = os.listdir('build/mo')
    for path in glob.glob(os.path.join(sys.prefix,
                                       "share/locale/*/LC_MESSAGES/{}.mo".format(name))):
        lang = os.path.split(os.path.split(os.path.split(path)[0])[0])[1]
        if lang in langs:
            addfile(os.path.relpath(path, sys.prefix))


addlocale("gtk30")


def addicons():
    addfile("share/icons/hicolor/index.theme")
    addfile("share/icons/Adwaita/index.theme")
    for i in ['places/folder', 'mimetypes/text-x-generic']:
        addfile(os.path.join('share/icons/Adwaita/16x16/', i + '.png'))
    icons = [
        'places/user-desktop',
        'places/user-home',
        'actions/bookmark-new',
        'actions/document-open-recent',
        'actions/folder-new',
        'actions/list-add',
        'actions/list-remove',
        'actions/media-eject',
        'actions/document-save',
        'actions/document-save-as',
        'actions/document-open',
        'actions/insert-image',
        'actions/object-rotate-left',
        'actions/object-rotate-right',
        'actions/open-menu',
        'actions/zoom-in',
        'actions/zoom-out',
        'actions/go-up',
        'actions/go-down',
        'ui/pan-down',
        'ui/pan-end',
        'ui/pan-start',
        'ui/pan-up',
        'ui/window-close',
        'ui/window-maximize',
        'ui/window-minimize',
        'ui/window-restore',
        'devices/drive-harddisk',
        'devices/drive-optical',
        'places/folder-documents',
        'places/folder-download',
        'places/folder-music',
        'places/folder-pictures',
        'places/folder-videos',
        'places/user-trash',
    ]

    for i in icons:
        addfile(os.path.join('share/icons/Adwaita/symbolic/', i + '-symbolic.svg'))

required_dlls = [
    'poppler-glib-8',
    'handy-1-0',
]

for dll in required_dlls:
    fn = 'lib' + dll + '.dll'
    include_files.append((os.path.join(sys.prefix, 'bin', fn), fn))

required_gi_namespaces = [
    "Gtk-3.0",
    "Gdk-3.0",
    "cairo-1.0",
    "Pango-1.0",
    "GObject-2.0",
    "GLib-2.0",
    "Gio-2.0",
    "GdkPixbuf-2.0",
    "GModule-2.0",
    "Atk-1.0",
    "Poppler-0.18",
    "HarfBuzz-0.0",
    "Handy-1",
    "freetype2-2.0",
]

for ns in required_gi_namespaces:
    addfile("lib/girepository-1.0/{}.typelib".format(ns))

addfile("lib/gdk-pixbuf-2.0/2.10.0/loaders/libpixbufloader-bmp.dll")
addfile("lib/gdk-pixbuf-2.0/2.10.0/loaders/pixbufloader_svg.dll")
addfile("lib/gdk-pixbuf-2.0/2.10.0/loaders/libpixbufloader-png.dll")
addfile("lib/gdk-pixbuf-2.0/2.10.0/loaders.cache")
addfile("share/glib-2.0/schemas/gschemas.compiled")
addicons()


# Add encoding files required to correctly display Chinese, Japanese, Korean
# and some Cyrillic characters in PDF documents. (from poppler-data package)
from_path = os.path.join(sys.prefix, 'share/poppler/')
to_path = 'lib/share/poppler/'
include_files.append((from_path, to_path))
addfile('share/licenses/poppler-data/COPYING')
addfile('share/licenses/poppler-data/COPYING.adobe')


# gspawn-helper is needed for website link in AboutDialog
from_path = os.path.join(sys.prefix, 'bin', 'gspawn-win64-helper.exe')
to_path = 'gspawn-win64-helper.exe'
include_files.append((from_path, to_path))


build_options = dict(
    packages=['gi', 'packaging', 'pikepdf'],
    excludes=['tkinter', 'test'],
    include_files=include_files,
)


def get_target_name(suffix):
    return 'pdfarranger-{}-windows-{}'.format(VERSION, suffix)


msi_options = dict(
    target_name=get_target_name('installer.msi'),
    upgrade_code='{ab1752a6-575c-42e1-a261-b85cb8a6b524}',
    extensions=[{
        "extension": "pdf",
        "verb": "open",
        "executable": "pdfarranger.exe",
        "argument": '"%1"',
    }]
)


class bdist_zip(distutils.cmd.Command):
    """ Minimalist command to create a Windows portable .zip distribution """
    description = "create a \"zip\" distribution"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        build_base = self.get_finalized_command('build').build_base
        build_exe = self.get_finalized_command('build_exe')
        fullname = self.distribution.get_fullname()
        build_exe.build_exe = os.path.join(build_base, fullname)
        build_exe.run()
        config_ini = os.path.join(build_exe.build_exe, 'config.ini')
        f = open(config_ini, 'w')
        f.close()
        dist_dir = self.get_finalized_command('bdist').dist_dir
        archname = os.path.join(dist_dir, get_target_name('portable'))
        self.make_archive(archname, 'zip', root_dir=build_base, base_dir=fullname)
        shutil.rmtree(build_exe.build_exe)


setup(name='PDF Arranger',
      author='The PDF Arranger team',
      version=VERSION,
      description='PDF Arranger',
      long_description='A simple application for PDF Merging, Rearranging, and Splitting',
      options=dict(build_exe=build_options, bdist_msi=msi_options),
      cmdclass={'bdist_zip': bdist_zip},
      packages=['pdfarranger'],
      executables=[Executable('pdfarranger/__main__.py',
                              base='Win32GUI' if sys.platform == 'win32' else None,
                              target_name='pdfarranger.exe',
                              icon='data/pdfarranger.ico',
                              shortcut_name='PDF Arranger',
                              shortcut_dir='StartMenuFolder'
                              )])



from dataclasses import dataclass, field
import os
import packaging.version as version
from typing import Any, List, Tuple
import unittest
from unittest.mock import Mock

import pikepdf

from pdfarranger.exporter import export
from pdfarranger.core import Dims, Sides


# The test files used for the tests in this file are in QDF format (see
# https://qpdf.readthedocs.io/en/stable/qdf.html).  They can be inspected with a simple text editor or compared with
# standard tools such as 'Meld'

def file(name):
    """Expand name to full filename"""
    return f'./tests/exporter/{name}.pdf'


@dataclass
class Page:
    """Mock Page class"""
    npage: int
    nfile: int = 1
    copyname: str = file('basic')
    angle: int = 0
    scale: float = 1.0
    crop: Sides = Sides()
    size_orig: Dims = Dims(612, 792)
    layerpages: List[Any] = field(default_factory=list)


@dataclass
class LayerPage:
    """Mock LayerPage class"""
    npage: int
    nfile: int = 1
    copyname: str = file('basic')
    angle: int = 0
    scale: float = 1.0
    crop: Sides = Sides()
    offset: Sides = Sides()
    laypos: str = 'OVERLAY'
    size_orig: Dims = Dims(612, 792)
    layerpages: List[Any] = field(default_factory=list)


class ExporterTest(unittest.TestCase):

    def compare_files(self, actual_file: str, expected_file: str) -> Tuple[bool, str]:
        """
        Compare two PDF files

        The following are ignored in the comparison:
        - blank lines
        - lines starting or ending with '%'
        - lines containing the name /Length
        - anything following the first '/' in line containing a Do operator
        - anything following the xref entry
        - lines starting with '/P'
        """
        with open(actual_file, 'rb') as f:
            actual = f.readlines()
        with open(expected_file, 'rb') as f:
            expected = f.readlines()

        actual_no = 0
        last_match = -1
        for line_no, line in enumerate(expected):
            if (len(line) == 0 or line.isspace() or line.startswith(b'%') or line.endswith(b'%\n')
                or b'/Length' in line or line.lstrip().startswith(b'/P')):
                pass
            elif line.startswith(b'xref'):
                return True, ''
            else:
                if b' Do ' in line or b' Do\n' in line:
                    line = line[:line.find(b'/')]

                try:
                    while not actual[actual_no].startswith(line):
                        actual_no += 1
                    last_match = actual_no
                    actual_no += 1
                except IndexError: # pragma: no cover
                    # Only get executed for failing test
                    print(line_no, line, last_match)
                    return (False, f'Failed to match line {line_no + 1} in expected  file {expected_file} : {line} \n'
                                   f'Last match in line {last_match + 1} of {actual_file}')

        return True, ''

    def case(self, test, files, *pages, start_with_empty=False):
        """Run a single test case."""
        if version.parse(pikepdf.__version__) < version.Version('8.0.0') or start_with_empty:
            expected_file = file(f'test{test}_out')
            start_with_empty = True
        else:
            expected_file = file(f'test{test}_8_out')
            if not os.path.exists(expected_file):
                expected_file = file(f'test{test}_out')

        mock_config = Mock()
        mock_config.start_with_empty.return_value = start_with_empty
        export(files, pages, {}, [file('out')], mock_config, None, True)
        self.assertTrue(*self.compare_files(file('out'), expected_file))

    def basic(self, test, *pages):
        """Test with basic.pdf as single input file."""
        self.case(test, [(file('basic'), '')], *pages)

    def outlines(self, test, *pages, ignore_8=None, start_with_empty=False):
        """Test with outlines.pdf as single input file."""
        if version.parse(pikepdf.__version__) >= version.Version('8.0.0'):
            self.case(test, [(file('outlines'), '')], *pages, start_with_empty=start_with_empty)

    def test01(self):
        """No transformations"""
        self.basic(1, Page(1), Page(2))

    def test02(self):
        """Rotate pages"""
        self.basic(2, Page(1, angle=90), Page(2, angle=180))

    def test03(self):
        """Rotate pages with existing /Rotate"""
        self.basic(3, Page(3, angle=90), Page(3, angle=180))

    def test04(self):
        """Scale page"""
        self.basic(4, Page(1, scale=0.25))

    def test05(self):
        """Crop page"""
        self.basic(5, Page(1, crop=Sides(0.1, 0.2, 0.3, 0.4)))

    def test06(self):
        """Rotate and crop page"""
        self.basic(6, Page(1, angle=90, crop=Sides(0.1, 0.2, 0.3, 0.4)),
                   Page(1, angle=180, crop=Sides(0.1, 0.2, 0.3, 0.4)),
                   Page(1, angle=270, crop=Sides(0.1, 0.2, 0.3, 0.4)))

    def test07(self):
        """Overlay page"""
        self.basic(7, Page(1, layerpages=[LayerPage(6)]))

    def test08(self):
        """Underlay page"""
        self.basic(8, Page(1, layerpages=[LayerPage(7, laypos='UNDERLAY')]))

    def test09(self):
        """Rotate overlay"""
        self.basic(9, Page(1, layerpages=[LayerPage(6, angle=90)]),
                   Page(1, layerpages=[LayerPage(6, angle=180)]))

    def test095(self):
        """Overlay page with itself - MediaBox with non-integer values"""
        self.case(95, [(file('overlay'), '')], Page(1, layerpages=[LayerPage(1)]))

    def test096(self):
        """Overlay page with itself - MediaBox with non-standard corners"""
        self.case(96, [(file('overlay'), '')], Page(2, layerpages=[LayerPage(2)]))

    def test10(self):
        """Offset overlay horizontal"""
        self.basic(10, Page(1, layerpages=[LayerPage(6, offset=Sides(.5, 0, 0, 0)),
                                           LayerPage(6, offset=Sides(0, 0.5, 0, 0))]))

    def test11(self):
        """Offset overlay vertical"""
        self.basic(11, Page(1, layerpages=[LayerPage(6, offset=Sides(0, 0, 0.5, 0)),
                                           LayerPage(6, offset=Sides(0, 0, 0, 0.5))]))

    def test12(self):
        """Duplicate page with annotations"""
        self.basic(12, Page(5), Page(5))

    def test13(self):
        """Rotate page with  annotations"""
        self.basic(13, Page(5, angle=90), Page(5, angle=180))

    def test14(self):
        """Scale page with annotations"""
        self.basic(14, Page(5, scale=0.25))

    def test15(self):
        """Crop page with annotations"""
        self.basic(15, Page(5, crop=Sides(0.05, 0.075, 0.2, 0.3)))

    def test16(self):
        """Overlay page with annotations"""
        self.basic(16, Page(1, layerpages=[LayerPage(5)]))

    def test17(self):
        """File with missing MediaBox"""
        self.case(17, [('./tests/test.pdf', '')], Page(1), Page(2))

    def test18(self):
        """Encrypted file"""
        self.case(18, [('./tests/test_encrypted.pdf', 'foobar')], Page(1), Page(2))

    def test19(self):
        """Copy file with outlines"""
        self.outlines(19, Page(1), Page(2), Page(3), Page(4))

    def test19a(self):
        """Copy file and discard outlines for pikepdf 8 """
        self.outlines(19, Page(1), Page(2), Page(3), Page(4), start_with_empty=True)

    def test20(self):
        """Reorder pages in file with outlines"""
        self.outlines(20, Page(4), Page(2), Page(1), Page(3))

    def test21(self):
        """Duplicate pages in file with outlines"""
        self.outlines(21, Page(4), Page(2), Page(3), Page(4))

    def test22(self):
        """Rotate pages in file with outlines"""
        self.outlines(22, Page(1), Page(2, angle=90), Page(3, angle=180), Page(4, angle=270))

    def test23(self):
        """Scale pages in file with outlines"""
        self.outlines(23, Page(1), Page(2, scale=0.5), Page(3), Page(4))

    def test24(self):
        """Import page with interactive form elements"""
        self.case(24, [(file('basic'), ''), (file('forms'), '')], Page(1), Page(1, nfile=2))

    def test25(self):
        """Duplicate page with interactive form elements"""
        self.case(25, [(file('forms'), '')], Page(1), Page(1))

    def test26(self):
        """Export with highest PDF versions of input files"""
        # gs -o exporter/test26_pdf1-4.pdf -sDEVICE=pdfwrite -g6120x7920 -dCompatibilityLevel=1.4 -c "showpage"
        # gs -o exporter/test26_pdf1-5.pdf -sDEVICE=pdfwrite -g6120x7920 -dCompatibilityLevel=1.5 -c "showpage"
        # to check versions
        # pdfinfo <pdf> | grep 'PDF version'
        self.case(
            26,
            [(file('test26_pdf1-4'), ''), (file('test26_pdf1-5'), '')],
            Page(1),
            Page(1, nfile=2),
        )



import configparser
import os
import subprocess
import sys
import unittest
import time
import tempfile
from typing import Tuple
import shutil
import packaging.version
from importlib import metadata
import warnings

import pikepdf

"""
Those tests are using Dogtail https://gitlab.com/dogtail/dogtail

Other tools using dogtail where you can get example on how to use it:

* https://gitlab.gnome.org/GNOME/gnome-boxes
* https://gitlab.gnome.org/GNOME/gnome-software
* https://gitlab.gnome.org/GNOME/gnome-characters
* https://gitlab.gnome.org/GNOME/eog
* https://github.com/kirienko/gourmet
* https://github.com/virt-manager/virt-manager
* https://gitlab.gnome.org/GNOME/gnome-weather/
* http://threedepict.sourceforge.net/
* https://salsa.debian.org/pkg-privacy-team/onioncircuits/-/tree/master/debian/tests

Example of Dogtail using a filechooser:

filechooser = app.child(roleName='file chooser')
treeview = filechooser.child(roleName='table', name='Files')
treeview.keyCombo('<ctrl>L')
treeview.typeText('qpdf-manual.pdf')
filechooser.button('Open').click()

You may need to run the following commands to run those tests in your current session instead of Xvfb:

* /usr/libexec/at-spi-bus-launcher --launch-immediately
* setxkbmap -v fr
* gsettings set org.gnome.desktop.interface toolkit-accessibility true

Tests need to be run with a window size of 545x600. It must be set in ~/.config/pdfarranger/config.ini.

Some tips:

* Use to print widget tree (names and roles) self._app().dump()

Example of how to run the test locally:
python3 -X tracemalloc -u -m unittest -v -f tests.test # run whole test
python3 -X tracemalloc -u -m unittest -v -f tests.test.TestBatch5 # run only TestBatch5
"""


def group(title):
    if "GITHUB_ACTIONS" in os.environ:
        print("::group::" + title)


def endgroup():
    if "GITHUB_ACTIONS" in os.environ:
        print("::endgroup::")


def check_img2pdf(version):
    try:
        import img2pdf
        v = [int(x) for x in img2pdf.__version__.split(".")]
        r = v >= version
    except Exception:
        r = False
    return r


def have_pikepdf8():
    return packaging.version.parse(
        metadata.version("pikepdf")
    ) >= packaging.version.Version("8.0")


class XvfbManager:
    """Base class for running offscreen tests"""

    def __init__(self, display=":99"):
        self.display = display
        env = os.environ.copy()
        env["DISPLAY"] = display
        self.xvfb_proc = None
        self.dbus_proc = None
        self.xvfb_proc = subprocess.Popen(["Xvfb", self.display])
        cmd = ["dbus-daemon", "--print-address=1", "--session"]
        self.dbus_proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, text=True, env=env
        )
        self.dbus_addr = self.dbus_proc.stdout.readline().strip()
        self.dbus_proc.stdout.close()
        os.environ["DISPLAY"] = display
        os.environ["DBUS_SESSION_BUS_ADDRESS"] = self.dbus_addr

    def kill(self):
        if "GITHUB_ACTIONS" not in os.environ:
            # Workaround. On GHA killing dbus also kill this python process.
            # And we don't care about zombie processes on GHA
            self.dbus_proc.kill()
            self.dbus_proc.wait()
        self.xvfb_proc.kill()
        self.xvfb_proc.wait()


class DogtailManager:
    def __init__(self):
        self.xvfb = None
        if "DISPLAY" not in os.environ:
            self.xvfb = XvfbManager()
        os.environ["LC_MESSAGES"] = "C"
        os.environ["GTK_MODULES"] = "gail:atk-bridge"
        cmd = "gsettings set org.gnome.desktop.interface toolkit-accessibility true"
        subprocess.check_call(cmd.split())
        # dogtail must be imported after setting DBUS_SESSION_BUS_ADDRESS
        from dogtail.config import config
        config.debugSleep = False
        # long duration at startup
        config.searchBackoffDuration = 1
        config.actionDelay = 0.01
        config.runInterval = 0.01
        config.defaultDelay = 0.1
        config.debugSearching = False
        config.searchCutoffCount = 30
        config.runTimeout = 1
        config.searchShowingOnly = True
        config.typingDelay = 0.05

    def kill(self):
        if self.xvfb is not None:
            self.xvfb.kill()


# dogtail does not support change of X11 server so it must be a singleton
dogtail_manager = DogtailManager()

def setUpModule():
    # if we can't kill the dbus process on GHA, let's at least suppress the warning
    if dogtail_manager.xvfb and "GITHUB_ACTIONS" in os.environ:
        warnings.filterwarnings("ignore",
                                f"subprocess {dogtail_manager.xvfb.dbus_proc.pid} is still running",
                                ResourceWarning)

def tearDownModule():
    dogtail_manager.kill()


class PdfArrangerManager:
    def __init__(self, args=None):
        self.process = None
        args = [] if args is None else args
        cmd = [sys.executable, "-u", "-X", "tracemalloc"]
        if "PDFARRANGER_COVERAGE" in os.environ:
            cmd = cmd + ["-m", "coverage", "run", "--concurrency=thread,multiprocessing", "-a"]
        self.process = subprocess.Popen(cmd + ["-m", "pdfarranger"] + args, stdout=subprocess.PIPE)

    def _get_stdout(self):
        """get the stdout of the pdfarranger process"""
        return self.process.stdout.read().decode("utf-8")

    def kill(self):
        self.process.stdout.close()
        self.process.kill()
        self.process.wait()

class PdfArrangerTest(unittest.TestCase):
    def _start(self, args=None, gtk_check=True):
        from dogtail.config import config
        config.searchBackoffDuration = 1
        self.__class__.pdfarranger = PdfArrangerManager(args)
        if not gtk_check:
            return
        # check that process is actually running
        self.assertIsNone(self._process().poll())
        self._app()
        # Now let's go faster
        config.searchBackoffDuration = 0.1

    def _get_stdout(self):
        """get the stdout of the pdfarranger process"""
        return self.__class__.pdfarranger._get_stdout()

    def _app(self):
        """Return the first instance of pdfarranger"""
        self._wait_cond(lambda: len(self._apps()) > 0)
        a =  self._apps()[0]
        self.assertFalse(a.dead)
        return a

    def _apps(self):
        """Return all instances of pdfarranger"""
        # Cannot import at top level because of DBUS_SESSION_BUS_ADDRESS
        from dogtail.tree import root

        return [
            a
            for a in root.applications()
            if a.name in ["__main__.py", "pdfarranger", "com.github.jeromerobert.pdfarranger"]
        ]

    def _mainmenu(self, action):
        mainmenu = self._app().child(roleName="toggle button", name="Menu")
        mainmenu.grabFocus()
        self._wait_cond(lambda: mainmenu.sensitive)
        mainmenu.click()
        if not isinstance(action, str):
            for submenu in action[:-1]:
                mainmenu.menu(submenu).point()
            action = action[-1]
        mainmenu.menuItem(action).click()

    def _wait_cond(self, cond):
        c = 0
        while not cond():
            time.sleep(0.1)
            self.assertLess(c, 300)
            c += 1

    def _find_by_role(self, role, node=None, show_only=False):
        if node is None:
            node = self._app()
        from dogtail import predicate

        return node.findChildren(
            predicate.GenericPredicate(roleName=role), showingOnly=show_only
        )

    def _is_saving(self):
        allstatusbar = self._find_by_role("status bar")
        statusbar = allstatusbar[-1]
        return statusbar.name.startswith("Saving")

    def _wait_saving(self):
        # When saving the main window is made unresponsive. When saving end
        # it's made responsive again. We must be sure that it's responsive
        # before continuing test else clicks may fail.
        self._wait_cond(lambda: not self._is_saving())

    def _status_text(self):
        self._app()
        allstatusbar = self._find_by_role("status bar")
        # If we have multiple status bar, consider the first one as the one who display the selection
        statusbar = allstatusbar[0]
        return statusbar.name

    def _assert_selected(self, selection):
        self.assertTrue(self._status_text().startswith("Selected pages: " + selection))

    def _assert_page_size(self, width, height, pageid=None):
        if pageid is not None:
            self._icons()[pageid].click()
            self._wait_cond(lambda: self._status_text().startswith(f"Selected pages: {pageid+1}"))
        label = " {:.1f} mm \u00D7 {:.1f} mm".format(width, height)
        self.assertTrue(self._status_text().endswith("Page Size:" + label))

    def _page_size(self, pageid):
        self._icons()[pageid].click()
        self._wait_cond(lambda: self._status_text().startswith(f"Selected pages: {pageid+1}"))
        text = self._status_text()
        size = text.split("Page Size: ")[1].split(" \u00D7 ")
        w, h = size[0].split(" mm")[0], size[1].split(" mm")[0]
        return float(w), float(h)

    def _check_file_content(self, filename, expected: Tuple[str]) -> Tuple[bool, str]:
        """
        Check expected is contained in file.
        """
        with open(filename, 'rb') as f:
            actual = f.readlines()

        n = 0
        for line in expected:
            try:
                while not actual[n].startswith(line):
                    n += 1
                n += 1
            except IndexError: # pragma: no cover
                # Only get executed for failing test
                return False, line
        return True, ''

    def _assert_page_content(self, filename: str, expected: Tuple[str]):
        """
        Check if a file in the current tmp folder contains expected content.
        """
        temp = os.path.join(self.__class__.tmp, 'temp.pdf')
        with pikepdf.Pdf.open(os.path.join(self.__class__.tmp, filename)) as pdf:
            pdf.save(temp, qdf=True, static_id=True, compress_streams=False,
                            stream_decode_level=pikepdf.StreamDecodeLevel.all)
        ok, content = self._check_file_content(temp, expected)
        self.assertTrue(ok, f'expectent content {content} missing in {filename}')

    def _assert_file_size(self, filename, size, tolerance=0.04):
        """
        Check if a file in the current tmp folder have the expected size
        at a given tolerance.
        """
        s = os.stat(os.path.join(self.__class__.tmp, filename))
        msg = "{} is {}b but is expected to be {} \u00B1 {}".format(
            filename, s.st_size, size, size * tolerance
        )
        self.assertLess(abs(s.st_size - size) / size, tolerance, msg=msg)

    def _icons(self):
        """Return the list of page icons"""
        viewport = self._app().child(roleName="layered pane")
        return self._find_by_role("icon", viewport)

    def _popupmenu(self, page, action):
        """Run an action on a give page using the popup menu"""
        if page is None:
            self._app().keyCombo("<shift>F10")
        else:
            self._icons()[page].click(button=3)
        popupmenu = self._app().child(roleName="window")
        if not isinstance(action, str):
            for submenu in action[:-1]:
                popupmenu.menu(submenu).point()
            action = action[-1]
        button = popupmenu.menuItem(action)
        self._wait_cond(lambda: button.sensitive)
        button.click()

    def _process(self):
        return self.__class__.pdfarranger.process

    def _import_file(self, filename, open_action=False):
        """Try to import a file with a file chooser and return that file chooser object"""
        self._mainmenu("Open" if open_action else "Import")
        filechooser = self._app().child(roleName='file chooser')
        treeview = filechooser.child(roleName="table", name="Files")
        treeview.keyCombo("<ctrl>L")
        treeview.typeText(os.path.abspath(filename))
        ob = filechooser.button("Open")
        self._wait_cond(lambda: ob.sensitive)
        ob.click()
        return filechooser

    def _save_as_chooser(self, filebasename, expected=None):
        """
        Fill and validate a Save As file chooser.
        The file chooser is supposed to be already open.
        """
        if expected is None:
            expected = [filebasename]
        filechooser = self._app().child(roleName="file chooser")
        tmp = self.__class__.tmp
        filename = os.path.join(tmp, filebasename)
        filechooser.child(roleName="text").text = filename
        saveb = filechooser.button("Save")
        self._wait_cond(lambda: saveb.sensitive)
        filechooser.button("Save").click()
        # Check files have been created
        for e in expected:
            self._wait_cond(lambda: os.path.isfile(os.path.join(tmp, e)))
        self._wait_cond(lambda: filechooser.dead)
        self._wait_saving()

    def _scale_selected(self, scale):
        app = self._app()
        app.keyCombo("S")
        dialog = app.child(roleName="dialog")
        relative = self._find_by_role("radio button", dialog)[1]
        relative.click()
        from dogtail import rawinput
        rawinput.keyCombo("Tab")
        rawinput.typeText(str(scale))
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    @staticmethod
    def _zoom(widget, n_events, zoom_in):
        """Zoom in/out with ctrl + mouse scroll wheel"""
        from dogtail import rawinput
        button = 4 if zoom_in else 5
        rawinput.holdKey("Control_L")
        for __ in range(n_events):
            widget.click(button=button)
        rawinput.releaseKey("Control_L")

    def _config(self):
        """Get config from config file used in test"""
        config = configparser.ConfigParser()
        config.read(self.__class__.config_file)
        return config

    def _quit(self):
        self._app().child(roleName="layered pane").grabFocus()
        self._app().child(roleName="layered pane").keyCombo("<ctrl>q")

    def _quit_without_saving(self):
        self._quit()
        dialog = self._app().child(roleName="alert")
        dialog.child(name="Don’t Save").click()
        # check that process actually exit
        self._process().wait(timeout=22)

    default_config = """
                     [preferences]
                     content-loss-warning = False
                     """

    @classmethod
    def setUpClass(cls):
        cls.pdfarranger = None
        cls.tmp = tempfile.mkdtemp()
        os.environ["APPDATA"] = cls.tmp
        os.makedirs(os.path.join(cls.tmp, "pdfarranger"))
        cls.config_file = os.path.join(os.environ["APPDATA"], "pdfarranger", "config.ini")
        config = configparser.ConfigParser()
        config.read_string(cls.default_config)
        if hasattr(cls, "test_config"):
            config.read_string(cls.test_config)
        with open(cls.config_file, "w") as f:
            config.write(f)

    def setUp(self):
        group("Running " + self.id())

    def tearDown(self):
        endgroup()

    @classmethod
    def tearDownClass(cls):
        if cls.pdfarranger:
            cls.pdfarranger.kill()
        if cls.tmp:
            shutil.rmtree(cls.tmp)


def with_config(test_config):
    """Decorator for test batch specific configuration settings

    test_config is the config.ini content to be applied on top of the default
    config defined in PdfArrangerTest.default_config
    """
    def f(cls):
        cls.test_config = test_config
        return cls
    return f


@with_config("""[preferences]
                content-loss-warning=True""")
class TestBatch1(PdfArrangerTest):
    def test_01_import_img(self):
        self._start(["data/screenshot.png"])
        # Handle the "content loss warning" dialog
        dialog = self._find_by_role("dialog")[-1]
        check_box = self._find_by_role("check box", dialog)[-1]
        check_box.click()  # Don't show this dialog again
        dialog.button("OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_02_properties(self):
        self._mainmenu("Edit Properties")
        dialog = self._app().child(roleName="dialog")
        creatorlab = dialog.child(roleName="table cell", name="Creator")
        creatorid = creatorlab.parent.children.index(creatorlab) + 1
        creatorval = creatorlab.parent.children[creatorid]
        creatorval.keyCombo("enter")
        from dogtail import rawinput
        rawinput.typeText('["Frodo", "Sam"]')
        dialog.child(name="OK").click()
        self._mainmenu("Edit Properties")
        dialog = self._app().child(roleName="dialog")
        rawinput.keyCombo("enter")
        rawinput.typeText('Memories')
        rawinput.keyCombo("enter")
	# FIXME: depending on where the test is ran the previous enter close
	# the dialog or do not close it.
        try:
            dialog.child(name="OK").click()
        except Exception:
            print("'Edit Properties dialog' closed by 'enter'.")
        self._wait_cond(lambda: dialog.dead)

    def test_03_zoom(self):
        app = self._app()
        zoomoutb = app.child(description="Zoom Out")
        zoominb = app.child(description="Zoom In")
        # maximum dezoom whatever the initial zoom level
        for _ in range(10):
            zoomoutb.click()
        for _ in range(3):
            zoominb.click()

    def test_04_rotate_undo(self):
        app = self._app()
        self._assert_selected("")
        app.keyCombo("<ctrl>a")  # select all
        self._assert_selected("1")
        app.keyCombo("<ctrl>Left")  # rotate left
        app.keyCombo("<ctrl>z")  # undo
        app.keyCombo("<ctrl>y")  # redo
        app.keyCombo("<ctrl>a")
        app.keyCombo("<ctrl>Right")  # rotate right
        app.keyCombo("<ctrl>Right")  # rotate right

    def test_05_duplicate(self):
        self._popupmenu(0, "Duplicate")
        app = self._app()
        self.assertEqual(len(self._icons()), 2)
        self._app().child(roleName="layered pane").grabFocus()
        app.keyCombo("<ctrl>a")
        app.keyCombo("<ctrl>c")
        for __ in range(3):
            app.keyCombo("<ctrl>v")
        self.assertEqual(len(self._icons()), 8)
        app.keyCombo("Right")
        app.keyCombo("Left")
        app.keyCombo("Down")
        self._assert_selected("4")
        app.keyCombo("Up")
        self._assert_selected("2")

    def test_06_crop_margins(self):
        self._popupmenu(None, ["Select", "Select Odd Pages"])
        self._assert_selected("1, 3, 5, 7")
        self._popupmenu(0, "Crop Margins…")
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="Show values").click()
        time.sleep(0.2)  # Avoid 'GTK_IS_RANGE (range)' failed
        croppanel = dialog.child(name="Crop Margins")
        cropbuttons = self._find_by_role("spin button", croppanel)
        for i in range(4):
            cropbuttons[i].click()
            cropbuttons[i].text = str((i+1)*4)
        dialog.child(name="OK").click()
        # TODO: find the condition which could replace this ugly sleep
        time.sleep(0.5)
        self._wait_cond(lambda: dialog.dead)

    def test_07_split_page(self):
        lp = self._app().child(roleName="layered pane")
        lp.grabFocus()
        lbefore = len(self._icons())
        self._popupmenu(0, ["Select", "Select Even Pages"])
        self._assert_selected("2, 4, 6, 8")
        self._mainmenu(["Edit", "Split Pages…"])
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)
        self.assertEqual(len(self._icons()), lbefore + 4)

    def test_08_zoom_pages(self):
        self._app().child(roleName="layered pane").keyCombo("Home")
        self._assert_selected("1")
        self._app().keyCombo("f")
        for __ in range(2):
            self._app().keyCombo("minus")
        # Zoom level is now 0 and that's what will be saved to config.ini and
        # used by next batches

    def test_09_save_as(self):
        self._mainmenu("Save")
        self._save_as_chooser("foobar.pdf")

    def test_10_reverse(self):
        self._popupmenu(0, ["Select", "Same Page Format"])
        self._assert_selected("1, 4, 7, 10")
        self._popupmenu(0, ["Select", "All From Same File"])
        self._assert_selected("1-12")
        self._popupmenu(0, "Reverse Order")

    def test_11_quit(self):
        self._mainmenu("Quit")
        dialog = self._app().child(roleName="alert")
        dialog.child(name="Cancel").click()
        self._app().keyCombo("<ctrl>s")
        self._wait_saving()
        self._quit()
        # check that process actually exit
        self._process().wait(timeout=22)
        self.assertEqual(self._config()["preferences"]["content-loss-warning"], "False")


class TestBatch2(PdfArrangerTest):
    def test_01_open_empty(self):
        self._start()

    def test_02_import(self):
        filechooser = self._import_file("tests/test.pdf")
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 2)

    def test_03_blank_page(self):
        self._popupmenu(0, ["Insert Blank Page…"])
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: len(self._icons()) == 3)

    def test_04_paste_and_resize_overlay(self):
        self._popupmenu(0, ["Copy"])
        self._popupmenu(1, ["Paste Special", "Paste As Overlay…"])
        dialog = self._app().child(roleName="dialog")
        da = dialog.child(roleName="drawing area")
        top_left = da.position[0] + 30, da.position[1] + 25
        center = da.position[0] + da.size[0] / 2, da.position[1] + da.size[1] / 2
        from dogtail import rawinput
        self._wait_cond(lambda: da.sensitive)
        rawinput.drag(top_left, center)
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_05_cropborder(self):
        self._popupmenu(1, "Crop White Borders")
        small_size = self._page_size(1)
        self._popupmenu(0, "Crop White Borders")
        large_size = self._page_size(0)
        width_ratio = small_size[0] / large_size[0]
        self.assertAlmostEqual(width_ratio, 0.5, places=1)

    def test_06_paste_and_move_overlay(self):
        app = self._app()
        app.keyCombo("<ctrl>c")
        app.keyCombo("End")
        app.keyCombo("<shift><ctrl>o")
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="Show values").click()
        time.sleep(0.2)  # Avoid 'GTK_IS_RANGE (range)' failed
        spinbtns = self._find_by_role("spin button", dialog)
        spinbtns[0].click()
        spinbtns[0].text = "10"
        spinbtns[1].click()
        spinbtns[1].text = "15"
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_07_past_underlay(self):
        """Past a page with overlay under an other page"""
        app = self._app()
        app.keyCombo("<ctrl>c")
        app.keyCombo("Home")
        app.keyCombo("<shift><ctrl>u")
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_08_clear(self):
        self._popupmenu(1, "Delete")
        self.assertEqual(len(self._icons()), 2)

    def test_09_export(self):
        self._mainmenu(["Export", "Export All Pages to Individual Files…"])
        self._save_as_chooser(
            "alltosingle.pdf", ["alltosingle.pdf", "alltosingle-002.pdf"]
        )
        self._assert_file_size("alltosingle.pdf", 1800)
        self._assert_file_size("alltosingle-002.pdf", 1544)
        self._assert_page_content("alltosingle.pdf", (
            b'1 0 0 rg 530 180 m 70 180 l 300 580 l h 530 180 m B',
            b'  /BBox [', b'    68\n', b'    179\n', b'    532\n', b'    582\n',
            b'1 0 0 rg 530 180 m 70 180 l 300 580 l h 530 180 m B'))
        self._assert_page_content("alltosingle.pdf", (
            b'  /BBox [', b'    0\n', b"    0\n", b'    612\n', b'    792\n',
            b'0 1 0 rg 530 180 m 70 180 l 300 580 l h 530 180 m B'))

    def test_10_about(self):
        self._mainmenu("About")
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="Close").click()
        self._wait_cond(lambda: dialog.dead)

    def test_11_quit(self):
        self._quit_without_saving()


class TestBatch3(PdfArrangerTest):
    """Test encryption"""
    def test_01_open_encrypted(self):
        filename = os.path.join(self.__class__.tmp, "other_encrypted.pdf")
        shutil.copyfile("tests/test_encrypted.pdf", filename)
        self._start([filename])
        dialog = self._app().child(roleName="dialog")
        passfield = dialog.child(roleName="password text")
        passfield.text = "foobar"
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_02_import_wrong_pass(self):
        filechooser = self._import_file("tests/test_encrypted.pdf")
        dialog = self._app().child(roleName="dialog")
        passfield = dialog.child(roleName="password text")
        passfield.text = "wrong"
        dialog.child(name="OK").click()
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="Cancel").click()
        self._wait_cond(lambda: dialog.dead)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 2)

    def test_03_quit(self):
        app = self._app()
        app.keyCombo("<ctrl>z")  # undo
        app.keyCombo("<ctrl>y")  # redo
        self._quit()
        dialog = self._app().child(roleName="alert")
        dialog.child(name="Save").click()
        filechooser = self._app().child(roleName="file chooser")
        filechooser.button("Save").click()
        dialog = self._app().child(roleName="alert")
        dialog.child(name="Replace").click()
        # check that process actually exit
        self._process().wait(timeout=22)


class TestBatch4(PdfArrangerTest):
    """Check the size of duplicated and scaled pages"""
    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_duplicate(self):
        app = self._app()
        app.keyCombo("Down")
        self._popupmenu(0, ["Duplicate"])
        app.keyCombo("Right")

    def test_03_scale(self):
        self._scale_selected(200)
        self._app().keyCombo("<ctrl>Left")  # rotate left
        self._assert_selected("2")
        self._assert_page_size(558.8, 431.8)

    def test_04_export(self):
        app = self._app()
        app.keyCombo("<ctrl>a")  # select all
        self._mainmenu(["Export", "Export Selection to a Single File…"])
        self._save_as_chooser("scaled.pdf")
        self._popupmenu(0, "Delete")

    def test_05_import(self):
        filename = os.path.join(self.__class__.tmp, "scaled.pdf")
        filechooser = self._import_file(filename)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 3)
        app = self._app()
        self._app().child(roleName="layered pane").grabFocus()
        app.keyCombo("Right")
        app.keyCombo("Right")
        self._assert_selected("2")
        self._assert_page_size(558.8, 431.8)
        self._quit_without_saving()


class TestBatch5(PdfArrangerTest):
    """Test booklet and blank pages"""
    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_blank_page(self):
        self._popupmenu(None, ["Select", "Select All"])
        self._popupmenu(0, ["Crop White Borders"])
        self._scale_selected(150)
        self._popupmenu(0, ["Insert Blank Page…"])
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: len(self._icons()) == 3)

    def test_03_booklet(self):
        self._popupmenu(0, ["Select", "Select All"])
        self._popupmenu(0, ["Booklet", "Generate (imposition)"])
        self._wait_cond(lambda: len(self._icons()) == 2)
        self._app().child(roleName="layered pane").keyCombo("Home")
        self._assert_page_size(491.1, 213.3)
        self._app().child(roleName="layered pane").keyCombo("End")
        self._assert_page_size(491.1, 213.3)

    def test_04_crop_white_border(self):
        # Test selection with shift+arrow
        self._app().child(roleName="layered pane").grabFocus()
        self._app().child(roleName="layered pane").keyCombo("<shift>Left")
        self._assert_selected("1-2")
        self._popupmenu(0, ["Crop White Borders"])
        self._assert_page_size(245.2, 213.3, 0)
        self._assert_page_size(245.2, 213.3, 1)

    def test_05_split_booklet(self):
        self._popupmenu(0, ["Select", "Select All"])
        self._popupmenu(0, ["Booklet", "Generate (imposition)"])
        self._popupmenu(0, ["Select", "Select All"])
        self._popupmenu(0, ["Booklet", "Split (unimposition)"])
        # OK OK hear me out generating the booklet took 2 pages and made it 2 after the process
        # by adding two blank pages at the end. So we should have 4 now.
        # They should have 245.2x213.3mm size like in test_04 because
        # test.pdf is only loaded once for all operations.
        self._wait_cond(lambda: len(self._icons()) == 4)
        self._app().child(roleName="layered pane").keyCombo("Home")
        self._assert_page_size(245.2, 213.3)
        self._app().child(roleName="layered pane").keyCombo("End")
        self._assert_page_size(245.2, 213.3)

    def test_06_buggy_exif(self):
        """Test img2pdf import with buggy EXIF rotation"""
        if not check_img2pdf([0,4,2]):
            print("Ignoring test_05_buggy_exif, img2pdf too old")
            return
        filechooser = self._import_file("tests/1x1.jpg")
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 5)

    def test_07_quit(self):
        self._quit_without_saving()


class TestBatch6(PdfArrangerTest):
    """Test hide margins and merge pages"""
    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_merge_pages(self):
        self._app().keyCombo("<ctrl>a")
        self._popupmenu(0, "Merge Pages…")
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_03_crop_margins(self):
        self._app().keyCombo("<ctrl>a")
        self._popupmenu(0, "Crop Margins…")
        dialog = self._app().child(roleName="dialog")
        croppanel = dialog.child(name="Crop Margins")
        cropbuttons = self._find_by_role("spin button", croppanel)
        dialog.child(name="Show values").click()
        for i in range(4):
            cropbuttons[i].typeText("2")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)
        self._assert_page_size(414.5, 268.2)

    def test_04_hide_margins(self):
        self._app().keyCombo("<ctrl>a")
        self._assert_selected("1")
        self._app().keyCombo("H")
        dialog = self._app().child(roleName="dialog")
        da = dialog.child(roleName="drawing area")
        page_x = da.position[0] + 25  # 25 = padding in DrawingAreaWidget
        page_width = da.size[0] - 50
        x_center = da.position[0] + da.size[0] / 2
        y_center = da.position[1] + da.size[1] / 2
        from dogtail import rawinput
        self._wait_cond(lambda: da.sensitive)
        for button in ["Apply", "Revert", "Apply"]:
            rawinput.drag((page_x, y_center), (page_x + page_width * 0.8, y_center))
            rawinput.drag((page_x + page_width * 0.9, y_center), (page_x + page_width * 0.3, y_center))
            dialog.child(name=button).click()
        hidepanel = dialog.child(name="Hide Margins")
        hidebuttons = self._find_by_role("spin button", hidepanel)
        hidebuttons[2].text = str(round(float(hidebuttons[2].text) / 10) * 10)
        hidebuttons[3].text = str(round(float(hidebuttons[3].text) / 10) * 10)
        self.assertEqual(hidebuttons[2].text, "60")
        self.assertEqual(hidebuttons[3].text, "20")
        self._zoom(da, 15, zoom_in=True)
        rawinput.drag((x_center, y_center), (x_center + 10, y_center + 20),  button=2)  # pan view
        self._zoom(da, 15, zoom_in=False)
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_05_export(self):
        self._popupmenu(0, ["Select", "Select All"])
        self._mainmenu(["Export", "Export Selection to a Single File…"])
        self._save_as_chooser("hide.pdf")
        self._assert_file_size("hide.pdf", 1726)

    def test_06_merge_pages(self):
        self._popupmenu(0, ["Select", "Select All"])
        self._popupmenu(0, "Merge Pages…")
        dialog = self._app().child(roleName="dialog")
        orderpanel = dialog.child(name="Page Order")
        radiobuttons = self._find_by_role("radio button", orderpanel)
        radiobuttons[0].click()
        radiobuttons[2].click()
        radiobuttons[4].click()
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)
        self._assert_page_size(829.1, 268.2)

    def test_07_quit(self):
        self._quit_without_saving()


class TestBatch7(PdfArrangerTest):
    """Test extracting of images and text"""
    def test_01_import_pdf(self):
        self._start(["tests/test_raster_image_text.pdf"])

    def test_02_explode_images(self):
        self._popupmenu(1, ["Select", "Select All"])
        all_selected = all([icon.selected for icon in self._icons()])
        self._wait_cond(lambda: all_selected)
        self._popupmenu(1, ["Extract", "Explode into Images"])
        self._wait_cond(lambda: len(self._icons()) == 5)

    def test_03_copy_and_paste_image(self):
        """Copy image from page which has one image"""
        self._icons()[0].click(button=1)
        self._popupmenu(0, ["Extract", "Copy Image"])
        self._app().keyCombo("<ctrl>v")
        self._wait_cond(lambda: len(self._icons()) == 6)

    def test_04_copy_and_paste_image(self):
        """Try to copy image from page which has two images"""
        self._popupmenu(2, ["Extract", "Copy Image"])
        filler = self._find_by_role("filler")[-1]
        filler.child(name="OK").click()
        self._wait_cond(lambda: filler.dead)
        self._popupmenu(1, ["Paste After"])
        self._wait_cond(lambda: len(self._icons()) == 7)

    def test_05_copy_text(self):
        """Copy text "tests/test_raster_image_text.pdf" from pdf"""
        self._popupmenu(0, ["Extract", "Copy Text"])
        self._app().keyCombo("<ctrl>v")
        self._wait_cond(lambda: len(self._icons()) == 9)

    def test_06_quit(self):
        self._quit_without_saving()


class TestBatch8(PdfArrangerTest):
    """Test Open action"""

    def test_01_open_empty(self):
        self._start()

    def test_02_open(self):
        filechooser = self._import_file("tests/test.pdf", open_action=True)
        self._wait_cond(lambda: filechooser.dead)

    def test_03_open_again(self):
        """Create a new pdfarranger instance"""
        self.assertEqual(len(self._apps()), 1)
        filechooser = self._import_file("tests/test.pdf", open_action=True)
        self._wait_cond(lambda: filechooser.dead)
        self._wait_cond(lambda: len(self._apps()) == 2)

    def test_04_quit(self):
        """Quit the second instance"""
        self._quit()
        self._wait_cond(lambda: len(self._apps()) == 1)

    def test_05_quit(self):
        """Quit the first instance"""
        self._quit()
        self._process().wait(timeout=22)

class TestBatch9(PdfArrangerTest):
    """test batch for cli flags """

    def test_01_cli_version(self):
        """test version flag"""
        self._start(["--version"], gtk_check=False)
        time.sleep(0.5)
        stdout = self._get_stdout()
        print(stdout)
        self.assertIn("pikepdf", stdout)
        self.assertIn("libqpdf", stdout)
        self.assertIn("OS-", stdout)
        self.assertIn("OS_Version-", stdout)


class TestBatch10(PdfArrangerTest):
    """Test start-with-empty config setting"""

    def test_01_default(self):
        """Test default setting"""
        self._start(["tests/test_encrypted.pdf"])
        dialog = self._app().child(roleName="dialog")
        passfield = dialog.child(roleName="password text")
        passfield.text = "foobar"
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

        self._mainmenu("Preferences")
        dialog = self._app().child(roleName="dialog")
        savepanel_name = "Saving/exporting to single file"
        if have_pikepdf8():
            savepanel = dialog.child(name=savepanel_name)
            checkboxes = self._find_by_role("check box", savepanel)
            self.assertTrue(checkboxes[0].checked)
        else:
            panels = self._find_by_role("panel", dialog)
            self.assertNotIn(savepanel_name, [p.name for p in panels])
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

        self._mainmenu("Save As…")
        self._save_as_chooser("encrypted.pdf", ["encrypted.pdf"] )
        if have_pikepdf8():
            with warnings.catch_warnings(record=True) as w:
                with pikepdf.open(os.path.join(self.__class__.tmp, "encrypted.pdf"), password="foobar"):
                    pass
            # there should be no warning that no password was needed to open this PDF
            self.assertEqual(len(w), 0)
        else:
            with pikepdf.open(os.path.join(self.__class__.tmp, "encrypted.pdf")):
                # there should be no pikepdf._core.PasswordError for pikepdf < 8
                pass

    def test_02_legacy(self):
        """Test legacy setting"""
        if have_pikepdf8():
            self._mainmenu("Preferences")
            dialog = self._app().child(roleName="dialog")
            savepanel = dialog.child(name="Saving/exporting to single file")
            checkboxes = self._find_by_role("check box", savepanel)
            checkboxes[0].click()
            dialog.child(name="OK").click()
            self.assertFalse(checkboxes[0].checked)
            self._wait_cond(lambda: dialog.dead)

        self._mainmenu("Save As…")
        self._save_as_chooser("unencrypted.pdf", ["unencrypted.pdf"])
        with pikepdf.open(os.path.join(self.__class__.tmp, "unencrypted.pdf")):
            # there should be no pikepdf._core.PasswordError
            pass

    def test_03_quit(self):
        self._quit()  # save config
        self._process().wait(timeout=22)
        if have_pikepdf8():
            self.assertEqual(self._config()["preferences"]["start-with-empty"], "True")


class TestBatch11(PdfArrangerTest):
    """Test show-save-warnings config setting and metadata"""

    def test_01_save_and_disable_warning(self):
        """Show warning discarded metadata (default)"""
        self._start(["tests/test_metadata1.pdf"])
        self._mainmenu("Save")
        self._save_as_chooser("output1.pdf")
        alert = self._find_by_role("alert")[0]
        from dogtail import predicate
        alert.findChild(predicate.IsNamed("Saving produced some warnings"), requireResult=True)
        check_box = self._find_by_role("check box", alert)[0]
        check_box.click()  # Don't show warnings when saving again
        alert.button("OK").click()
        self._wait_cond(lambda: alert.dead)

    def test_02_save_without_warning(self):
        """Don't show warning discarded metadata"""
        self._mainmenu("Save")
        self._wait_saving()

    def test_03_retain_metadata(self):
        """Merge metadata into first file and don't overwrite from info dict"""
        filechooser = self._import_file("tests/test_metadata2.pdf")
        self._wait_cond(lambda: filechooser.dead)
        self._mainmenu("Save As…")
        self._save_as_chooser("output2.pdf", ["output2.pdf"])
        with pikepdf.Pdf.open(os.path.join(self.__class__.tmp, "output2.pdf")) as pdf:
            del pdf.docinfo["/Producer"]  # pikepdf version will vary
            expected = {
                "/Title": "Title1",
                "/Author": "First; Second",
                "/Subject": "Subject2",
                "/Custom": "info dict entry",
            }
            if not have_pikepdf8():
                del expected["/Custom"]
            self.assertDictEqual(dict(pdf.docinfo), expected)
            with pdf.open_metadata(False, False) as meta:
                del meta["pdf:Producer"]      # pikepdf version will vary
                del meta["xmp:MetadataDate"]  # will vary
                exptected = {
                    # xmp metadata with info dict equivalents
                    "{http://purl.org/dc/elements/1.1/}title": "Title1",
                    "{http://purl.org/dc/elements/1.1/}creator": ["First", "Second"],
                    "{http://purl.org/dc/elements/1.1/}description": "Subject2",
                    # xmp metadata without info dict equivalents
                    "{http://purl.org/dc/elements/1.1/}identifier": "ID1",
                    "{http://purl.org/dc/elements/1.1/}source": "Source2",
                    "{http://ns.adobe.com/xap/1.0/}label": "Label1",
                }
                self.assertDictEqual(dict(meta), exptected)

    def test_04_quit(self):
        """Test config settings persistence"""
        self._quit()
        self._process().wait(timeout=22)
        self.assertEqual(self._config()["preferences"]["show-save-warnings"], "False")


class TestBatch12(PdfArrangerTest):
    """Test printing"""

    def test_01_open_documents(self):
        self._start(["data/screenshot.png", "tests/test.pdf", "tests/test_metadata1.pdf"])
        # Remove last page
        self._popupmenu(3, ["Cut"])

    def test_02_print_all_to_pdf(self):
        """Print with Gtk.PrintUnixDialog"""
        self._mainmenu(["Print…"])
        print_dialog = self._app().child(roleName="dialog")
        print_dialog.grabFocus()
        print_to_file = print_dialog.child(name="Print to File")
        print_to_file.grabFocus()
        print_to_file.click()
        from dogtail import rawinput
        rawinput.keyCombo("Tab")
        # Open "file chooser" to set the output name
        rawinput.keyCombo("enter")
        filechooser = self._app().child(roleName="file chooser")
        output = os.path.join(self.__class__.tmp, 'output.pdf')
        filechooser.child(roleName="text").text = output
        filechooser.button("Select").click()
        self._wait_cond(lambda: filechooser.dead)
        print_dialog.button("Print").click()
        self._wait_cond(lambda: print_dialog.dead)
        # Import the printed pdf
        filechooser = self._import_file(output)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 6)

    def test_03_quit(self):
        self._quit_without_saving()


class TestBatch13(PdfArrangerTest):
    """Test export as jpg images"""

    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_export_as_png(self):
        """Export image at default 150 pixels/inch"""
        self._popupmenu(None, ["Select", "Select All"])
        self._mainmenu(["Export", "Export Selection to JPG Images…"])
        self._save_as_chooser("image-001.jpg")

    def test_03_import_png(self):
        """Import the two images that were created"""
        tmp = self.__class__.tmp
        filename = os.path.join(tmp, "image-001.jpg")
        filechooser = self._import_file(filename)
        self._wait_cond(lambda: filechooser.dead)
        filename = os.path.join(tmp, "image-002.jpg")
        filechooser = self._import_file(filename)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 4)
        self._app().child(roleName="layered pane").grabFocus()
        self._app().keyCombo("End")
        self._assert_page_size(215.9, 279.4)

    def test_04_quit(self):
        self._quit_without_saving()


class TestBatch14(PdfArrangerTest):
    """Test export as png images"""

    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_export_as_png(self):
        """Export image at default 150 pixels/inch"""
        self._popupmenu(None, ["Select", "Select All"])
        self._mainmenu(["Export", "Export Selection to PNG Images…"])
        self._save_as_chooser("image-001.png")

    def test_03_close(self):
        """Test close application"""
        self._app().child(roleName="layered pane").grabFocus()
        self._app().keyCombo("<ctrl>W")
        self.assertEqual(len(self._icons()), 0)

    def test_04_import_png(self):
        """Import the two images that were created"""
        tmp = self.__class__.tmp
        filename1 = os.path.join(tmp, "image-001.png")
        filename2 = os.path.join(tmp, "image-002.png")
        filechooser = self._import_file(filename1)
        self._wait_cond(lambda: filechooser.dead)
        filechooser = self._import_file(filename2)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 2)
        self._app().child(roleName="layered pane").grabFocus()
        self._app().keyCombo("End")
        self._assert_page_size(215.9, 279.4)

    def test_05_quit(self):
        self._quit_without_saving()


class TestBatch15(PdfArrangerTest):
    """Test export as Rasterized PDF"""

    def test_01_import_pdf(self):
        self._start(["tests/test.pdf"])

    def test_02_export_as_pdf(self):
        """Export image at 150 pixels/inch as optimized greyscale"""
        self._mainmenu("Preferences")
        dialog = self._app().child(roleName="dialog")
        imagepanel = dialog.child(name="Image Export")
        checkboxes = self._find_by_role("check box", imagepanel)
        checkboxes[0].click()  # Greyscale
        checkboxes[1].click()  # Optimize
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)
        self._popupmenu(1, ["Rotate Left"])
        self._app().child(roleName="layered pane").grabFocus()
        self._popupmenu(None, ["Select", "Select All"])
        self._mainmenu(["Export", "Export Selection to Rasterized PDF (png)…"])
        self._save_as_chooser("pdf-image-001.pdf")

    def test_03_import_png(self):
        """Import the pdf that was created"""
        tmp = self.__class__.tmp
        filename = os.path.join(tmp, "pdf-image-001.pdf")
        filechooser = self._import_file(filename)
        self._wait_cond(lambda: filechooser.dead)
        self.assertEqual(len(self._icons()), 4)
        self._app().child(roleName="layered pane").grabFocus()
        self._app().keyCombo("End")
        self._assert_page_size(279.4, 215.9)

    def test_04_quit(self):
        self._quit_without_saving()


class TestBatch16(PdfArrangerTest):
    """Test find text"""

    def test_01_import_pdf(self):
        self._start(["tests/test_raster_image_text.pdf"])

    def test_02_split_page(self):
        self._popupmenu(0, ["Split Pages…"])
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)
        self.assertEqual(len(self._icons()), 3)

    def test_03_paste_as_overlay(self):
        self._popupmenu(0, ["Copy"])
        self._popupmenu(2, ["Paste Special", "Paste As Overlay…"])
        dialog = self._app().child(roleName="dialog")
        dialog.child(name="OK").click()
        self._wait_cond(lambda: dialog.dead)

    def test_04_find_prev(self):
        from dogtail import rawinput
        rawinput.keyCombo("<ctrl>F")
        rawinput.typeText("tests")
        rawinput.keyCombo("<shift>F3")
        time.sleep(0.3)
        self._assert_selected("3")

    def test_05_find_next(self):
        from dogtail import rawinput
        rawinput.keyCombo("<ctrl>A")
        rawinput.typeText("tes")
        rawinput.keyCombo("enter")
        time.sleep(0.3)
        rawinput.keyCombo("F3")
        time.sleep(0.3)
        self._mainmenu(["Find", "Find Next"])
        time.sleep(0.3)
        self._assert_selected("1")

    def test_06_find_all(self):
        from dogtail import rawinput
        rawinput.keyCombo("<ctrl>A")
        rawinput.typeText("te")
        self._mainmenu(["Find", "Find All"])
        time.sleep(0.3)
        self._assert_selected("1-3")

    def test_07_quit(self):
        from dogtail import rawinput
        rawinput.keyCombo("esc")
        self._quit_without_saving()



# Copyright (C) 2025 pdfarranger contributors
#
# pdfarranger is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import Gtk, Gdk
import gettext

from .core import Sides
from .exporter import get_in_memory_poppler_doc

_ = gettext.gettext


class SearchBarWidget(Gtk.SearchBar):
    """A widget for searching of text in PDF"""
    def __init__(self, window, iconview, pdfqueue, show_find_results, clear_find_results):
        super().__init__(show_close_button=True)
        self.button_box = Gtk.Box(homogeneous=True, halign=Gtk.Align.START, margin_start=6)

        button_prev = Gtk.Button.new_from_icon_name('go-up-symbolic', True)
        button_prev.set_tooltip_text(_("Find Previous"))
        button_prev.connect('clicked', self.find_sequent, -1)
        self.button_box.pack_start(button_prev, True, True, 0)

        button_next = Gtk.Button.new_from_icon_name('go-down-symbolic', True)
        button_next.set_tooltip_text(_("Find Next"))
        button_next.connect('clicked', self.find_sequent, 1)
        self.button_box.pack_start(button_next, True, True, 0)

        button_all = Gtk.Button(label=_("All"))
        button_all.set_tooltip_text(_("Find All"))
        button_all.connect('clicked', self.find_all)
        self.button_box.pack_start(button_all, True, True, 0)

        self.entry = Gtk.SearchEntry(width_chars=32)
        self.entry.connect('search_changed', self.enable_actions)
        entry_tools = Gtk.Box()
        entry_tools.pack_start(self.entry, True, True, 0)
        entry_tools.pack_start(self.button_box, True, True, 0)
        entry_tools.connect('unmap', self.close)

        self.add(entry_tools)
        self.connect_entry(self.entry)
        self.window = window
        self.iconview = iconview
        self.pdfqueue = pdfqueue
        self.model = self.iconview.get_model()
        self.show_find_results = show_find_results
        self.clear_find_results = clear_find_results
        self.text_old = ""
        self.page = None
        self.npage = 0
        self.rectangles = []
        self.nrect = 0
        self.event_cnt = 0

    def reveal(self):
        if not self.get_search_mode():
            self.entry.set_text(self.text_old)
            self.entry.grab_focus()
            self.set_search_mode(True)
        else:
            self.entry.grab_focus_without_selecting()

    def handle_event(self, event):
        """Forward key press events to entry"""
        handled = False
        if self.get_search_mode():
            if event.keyval == Gdk.KEY_Escape:
                handled = True
                self.set_search_mode(False)
            elif self.entry.has_focus():
                handled = True
                if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_x:
                    self.entry.cut_clipboard()
                elif event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_c:
                    self.entry.copy_clipboard()
                elif event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_v:
                    self.entry.paste_clipboard()
                elif event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_a:
                    self.entry.grab_focus()
                elif event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
                    self.find_sequent(step=1)
                else:
                    handled = self.entry.handle_event(event)
        return handled

    def enable_actions(self, _widget=None):
        enable = (self.entry.get_text() and self.get_search_mode() or
                  self.text_old and not self.get_search_mode())
        self.button_box.set_sensitive(enable)
        self.window.lookup_action("find_prev").set_enabled(enable)
        self.window.lookup_action("find_next").set_enabled(enable)
        self.window.lookup_action("find_all").set_enabled(enable)

    def find(self, _action, _option, _unknown):
        self.reveal()

    def find_prev(self, _action, _option, _unknown):
        self.find_sequent(step=-1)

    def find_next(self, _action, _option, _unknown):
        self.find_sequent(step=1)

    def find_all(self, _widget, _option=None, _unknown=None):
        self.event_cnt += 1
        event_cnt = self.event_cnt
        if len(self.model) == 0:
            return
        self.reveal()
        self.clear_find_results(unselect_all=True)
        text = self.entry.get_text()
        if text == "":
            return
        for npage in range(len(self.model)):
            rectangles = self.find_text(npage, text)
            if len(rectangles) > 0:
                self.show_find_results(npage, rectangles)
            self.entry.set_progress_fraction((npage + 1) / max(1, len(self.model)))
            if self.abort(event_cnt):
                break
        self.entry.set_progress_fraction(1)
        self.entry.set_progress_fraction(0)

    def find_sequent(self, _widget=None, step=1):
        self.event_cnt += 1
        event_cnt = self.event_cnt
        if len(self.model) == 0:
            return
        self.reveal()
        selection = self.iconview.get_selected_items()
        if len(selection) > 0:
            self.npage = selection[-1].get_indices()[0]
        self.npage = min(self.npage, len(self.model) - 1)
        self.clear_find_results(unselect_all=True)
        text = self.entry.get_text()
        if text == "":
            return
        text_changed = text != self.text_old
        page_changed = self.model[self.npage][0].__repr__() != self.page.__repr__()
        if text_changed or page_changed:
            # Search in current page
            self.rectangles = self.find_text(self.npage, text)
            self.nrect = -1 if step == 1 else len(self.rectangles)

        if 0 <= self.nrect + step < len(self.rectangles):
            # Get next rectangle index
            self.nrect += step
        else:
            # Continue searching until text is found or all has been searched
            self.rectangles = []
            cnt = 0
            while len(self.rectangles) == 0 and cnt < len(self.model):
                self.npage = (self.npage + step) % len(self.model)
                self.rectangles = self.find_text(self.npage, text)
                cnt += 1
                self.entry.set_progress_fraction(cnt / max(1, len(self.model)))
                if self.abort(event_cnt):
                    break
            self.nrect = 0 if step == 1 else len(self.rectangles) - 1

        if len(self.rectangles) > 0 and self.event_cnt == event_cnt:
            rectangle = self.rectangles[self.nrect]
            self.show_find_results(self.npage, [rectangle])
        self.entry.set_progress_fraction(1)
        self.entry.set_progress_fraction(0)

    def find_text(self, npage, text):
        if len(self.model) == 0:
            return []
        npage = min(npage, len(self.model) - 1)
        self.page = self.model[npage][0].duplicate(incl_thumbnail=False)
        page = self.page.duplicate(incl_thumbnail=False)
        page.rotate(-page.angle)
        crop = page.crop
        if len(page.layerpages) == 0:
            doc = self.pdfqueue[page.nfile - 1]
            poppler_page = doc.get_page(page.npage - 1)
        else:
            # Page has layers -> for simplicity export and search in exported pdf.
            # This is slower but easier than recalculating rectangle coordinates.
            rescale = 1 / page.scale
            page.scale = 1
            for lp in page.layerpages:
                lp.scale *= rescale
            page.crop = Sides()
            page.hide = Sides()
            doc, _buf = get_in_memory_poppler_doc([page], self.pdfqueue)
            poppler_page = doc.get_page(0)
        rectangles = poppler_page.find_text(text)
        rectangles = self.apply_crop(rectangles, poppler_page.get_size(), crop)
        if text != "":
            self.text_old = text
        return rectangles

    def apply_crop(self, rectangles, size, crop):
        for r in rectangles:
            r.x1 -= size[0] * crop.left
            r.x2 -= size[0] * crop.left
            r.y1 -= size[1] * crop.bottom
            r.y2 -= size[1] * crop.bottom
        rectangles = self.visible_rectangles(rectangles, size, crop)
        return rectangles

    def visible_rectangles(self, rectangles, size, crop):
        """Remove rectangles which are cropped away"""
        to_remove = []
        for num, r in enumerate(rectangles):
            if (r.x1 > size[0] * (1 - crop.left - crop.right) or
                r.x2 < 0 or
                r.y1 > size[1] * (1 - crop.top - crop.bottom) or
                r.y2 < 0
                ):
                to_remove.append(num)
        for num in reversed(to_remove):
            del rectangles[num]
        return rectangles

    def abort(self, event_cnt):
        while Gtk.events_pending():
            Gtk.main_iteration()
        closed = not self.get_search_mode()
        cleared = not self.entry.get_text()
        old_event = self.event_cnt > event_cnt
        return closed or cleared or old_event

    def close(self, _widget):
        self.clear_find_results(unselect_all=False)



import doctest
import unittest

import pdfarranger.core as core


class PTest(unittest.TestCase):
    """Base class for Page and LayerPage tests"""

    @staticmethod
    def _lpage1() -> core.LayerPage:
        """Sample layer page 1"""
        return core.LayerPage(2, 4, 'lcopy', 90, 2, core.Sides(0.11, 0.21, 0.31, 0.41),
                              core.Sides(0.12, 0.22, 0.32, 0.42), 'OVERLAY', core.Dims(10.33, 20.33))

    @staticmethod
    def _lpage1_90() -> core.LayerPage:
        """Sample layer page 1 rotated 90 degrees"""
        return core.LayerPage(2, 4, 'lcopy', 180, 2, core.Sides(0.41, 0.31, 0.11, 0.21),
                              core.Sides(0.42, 0.32, 0.12, 0.22), 'OVERLAY', core.Dims(10.33, 20.33))

    @staticmethod
    def _lpage1_180() -> core.LayerPage:
        """Sample layer page 1 rotated 180 degrees"""
        return core.LayerPage(2, 4, 'lcopy', 270, 2, core.Sides(0.21, 0.11, 0.41, 0.31),
                              core.Sides(0.22, 0.12, 0.42, 0.32), 'OVERLAY', core.Dims(10.33, 20.33))

    @staticmethod
    def _lpage1_270() -> core.LayerPage:
        """Sample layer page 1 rotated 180 degrees"""
        return core.LayerPage(2, 4, 'lcopy', 0, 2, core.Sides(0.31, 0.41, 0.21, 0.11),
                              core.Sides(0.32, 0.42, 0.22, 0.12), 'OVERLAY', core.Dims(10.33, 20.33))

    def _page1(self) -> core.Page:
        """Sample page 1"""
        return core.Page(1, 2, 0.55, 'copy', 0, 2, core.Sides(0.1, 0.2, 0.3, 0.4),
                         core.Sides(0.11, 0.21, 0.31, 0.41), core.Dims(100.33, 200.66), 'base', [self._lpage1()])

    def _page1_90(self) -> core.Page:
        """Sample page 1 rotated 90 degrees"""
        return core.Page(1, 2, 0.55, 'copy', 90, 2, core.Sides(0.4, 0.3, 0.1, 0.2),
                         core.Sides(0.41, 0.31, 0.11, 0.21), core.Dims(100.33, 200.66), 'base', [self._lpage1_90()])

    def _page1_180(self) -> core.Page:
        """Sample page 1 rotated 90 degrees"""
        return core.Page(1, 2, 0.55, 'copy', 180, 2, core.Sides(0.2, 0.1, 0.4, 0.3),
                         core.Sides(0.21, 0.11, 0.41, 0.31), core.Dims(100.33, 200.66), 'base', [self._lpage1_180()])

    def _page1_270(self) -> core.Page:
        """Sample page 1 rotated 90 degrees"""
        return core.Page(1, 2, 0.55, 'copy', 270, 2, core.Sides(0.3, 0.4, 0.2, 0.1),
                         core.Sides(0.31, 0.41, 0.21, 0.11), core.Dims(100.33, 200.66), 'base', [self._lpage1_270()])


class BasePageTest(PTest):

    def test01(self):
        """Test width | height | size_in_points"""
        self.assertAlmostEqual(self._page1().size_in_points()[0], 140.462)
        self.assertAlmostEqual(self._page1().width_in_points(), 140.462)
        self.assertAlmostEqual(self._page1().size_in_points()[1], 120.396)
        self.assertAlmostEqual(self._page1().height_in_points(), 120.396)
        self.assertAlmostEqual(self._page1_90().size_in_points()[0], 120.396)
        self.assertAlmostEqual(self._page1_90().width_in_points(), 120.396)
        self.assertAlmostEqual(self._page1_90().size_in_points()[1], 140.462)
        self.assertAlmostEqual(self._page1_90().height_in_points(), 140.462)

    def test02(self):
        """Test rotate_times"""
        #  Remember - counter-clockwise !
        self.assertEqual(core.Page.rotate_times(0), 0)
        self.assertEqual(core.Page.rotate_times(90), 3)
        self.assertEqual(core.Page.rotate_times(134), 3)
        self.assertEqual(core.Page.rotate_times(-270), 3)
        self.assertEqual(core.Page.rotate_times(3690), 3)


class PageTest(PTest):

    def _rotate(self, angle: int) -> core.Page:
        """Return sample page 1 rotated by angle"""
        p = self._page1()
        p.rotate(angle)
        return p

    def test01(self):
        """Test rotate"""
        self.assertEqual(repr(self._rotate(0)), repr(self._page1()))
        self.assertEqual(repr(self._rotate(3600)), repr(self._page1()))
        self.assertEqual(repr(self._rotate(-7200)), repr(self._page1()))
        self.assertEqual(repr(self._rotate(90)), repr(self._page1_90()))
        self.assertEqual(repr(self._rotate(-270)), repr(self._page1_90()))
        self.assertEqual(repr(self._rotate(180)), repr(self._page1_180()))
        self.assertEqual(repr(self._rotate(270)), repr(self._page1_270()))
        self.assertEqual(self._rotate(0).size, core.Dims(100.33, 200.66))
        self.assertEqual(self._rotate(90).size, core.Dims(200.66, 100.33))
        self.assertEqual(self._rotate(180).size, core.Dims(100.33, 200.66))
        self.assertEqual(self._rotate(270).size, core.Dims(200.66, 100.33))

    def test02(self):
        """Test duplicate"""
        p = self._page1()
        d = p.duplicate()
        self.assertEqual(repr(p), repr(d))
        p.rotate(90)
        self.assertEqual(repr(d), repr(self._page1()))
        self.assertNotEqual(repr(p), repr(self._page1()))

    def test03(self):
        """Test serialize"""
        self.assertEqual(self._page1().serialize(),
                         'copy///2///base///0///2///0.1///0.2///0.3///0.4///0.11///0.21///0.31///0.41///'
                         'lcopy///4///90///2///OVERLAY///0.11///0.21///0.31///0.41///0.12///0.22///0.32///0.42')

    def test04(self):
        """Test width | height | size_in_pixel"""
        self.assertEqual(self._page1().size_in_pixel()[0], 77)
        self.assertEqual(self._page1().width_in_pixel(), 77)
        self.assertEqual(self._page1().size_in_pixel()[1], 66)
        self.assertEqual(self._page1().height_in_pixel(), 66)
        self.assertEqual(self._page1_90().size_in_pixel()[0], 66)
        self.assertEqual(self._page1_90().width_in_pixel(), 66)
        self.assertEqual(self._page1_90().size_in_pixel()[1], 77)
        self.assertEqual(self._page1_90().height_in_pixel(), 77)
        self.assertTrue(isinstance(self._page1().height_in_pixel(), int), 'height_in_pixel not an int')
        self.assertTrue(isinstance(self._page1().width_in_pixel(), int), 'width_in_pixel not an int')


class LayerPageTest(PTest):

    def _rotate(self, angle: int) -> core.LayerPage:
        """Return sample layer page 1 rotated by angle"""
        p = self._lpage1()
        p.rotate(angle)
        return p

    def test01(self):
        """Test rotate"""
        self.assertEqual(repr(self._rotate(0)), repr(self._lpage1()))
        self.assertEqual(repr(self._rotate(-40)), repr(self._lpage1()))
        self.assertEqual(repr(self._rotate(80)), repr(self._lpage1()))
        self.assertEqual(repr(self._rotate(-1)), repr(self._lpage1_90()))
        self.assertEqual(repr(self._rotate(3)), repr(self._lpage1_90()))
        self.assertEqual(repr(self._rotate(-2)), repr(self._lpage1_180()))
        self.assertEqual(repr(self._rotate(-3)), repr(self._lpage1_270()))
        self.assertEqual(self._rotate(0).size, core.Dims(20.33, 10.33))
        self.assertEqual(self._rotate(-1).size, core.Dims(10.33, 20.33))
        self.assertEqual(self._rotate(-2).size, core.Dims(20.33, 10.33))
        self.assertEqual(self._rotate(-3).size, core.Dims(10.33, 20.33))

    def test02(self):
        """Test duplicate"""
        p = self._lpage1()
        d = p.duplicate()
        self.assertEqual(repr(p), repr(d))
        p.rotate(90)
        self.assertEqual(repr(d), repr(self._lpage1()))
        self.assertNotEqual(repr(p), repr(self._lpage1()))

    def test03(self):
        """Test serialize"""
        self.assertEqual(self._lpage1().serialize(),
                         'lcopy///4///90///2///OVERLAY///0.11///0.21///0.31///0.41///0.12///0.22///0.32///0.42')


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(core))
    return tests



#!/usr/bin/env python3

#
# pdfarranger - GTK+ based utility for splitting, rearrangement and
# modification of PDF documents.
# Copyright (C) 2008-2017 Konstantinos Poulios
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from setuptools import Command
from setuptools import setup
from setuptools import __version__ as setuptools_version

# support distros that ship old setuptools
setuptools_version = tuple(int(n) for n in setuptools_version.split('.')[:2])
if setuptools_version < (65, 2):
    from distutils.command.build import build
else:
    from setuptools.command.build import build

from os.path import join
import glob
import os
import subprocess

data_files = [
    ('share/applications', ['data/com.github.jeromerobert.pdfarranger.desktop']),
    ('share/pdfarranger', ['data/pdfarranger.ui', 'data/menu.ui']),
    ('share/man/man1', ['doc/pdfarranger.1']),
    ('share/metainfo', ['data/com.github.jeromerobert.pdfarranger.metainfo.xml']),
]


def _dir_to_data_files(src_dir, target_dir):
    data_files = []
    for root, _, files in os.walk(src_dir):
        tgt = join(target_dir, os.path.relpath(root, src_dir))
        if files:
            data_files.append((tgt, [join(root, f) for f in files]))
    return data_files


def _data_files(command):
    """Return the data_files of a command"""
    data_files = command.distribution.data_files
    if data_files is None:
        data_files = []
        command.distribution.data_files = data_files
    return data_files


class build_i18n(Command):
    description = "Build gettext .mo files"

    def initialize_options(self):
        self.build_base = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_base", "build_base"))

    def run(self):
        mo_dir = join(self.build_base, "mo")
        for filename in glob.glob(join("po", "*.po")):
            lang = os.path.basename(filename)[:-3]
            lang_dir = join(self.build_base, "mo", lang, "LC_MESSAGES")
            os.makedirs(lang_dir, exist_ok=True)
            subprocess.check_call(
                ["msgfmt", filename, "-o", join(lang_dir, "pdfarranger.mo")]
            )
        data_files = _data_files(self)
        data_files += _dir_to_data_files(mo_dir, join("share", "locale"))


class build_icons(Command):
    description = "Ensure icons get installed"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        src_icons = join("data", "icons")
        tgt_icons = join("share", "icons")
        data_files = _data_files(self)
        data_files += _dir_to_data_files(src_icons, tgt_icons)


build.sub_commands += [(x, lambda _: True) for x in ["build_i18n", "build_icons"]]

setup(
    name='pdfarranger',
    version='1.13.0',
    author='Jerome Robert',
    author_email='jeromerobert@gmx.com',
    description='A simple application for PDF Merging, Rearranging, and Splitting',
    url='https://github.com/pdfarranger/pdfarranger',
    license='GNU GPL-3',
    packages=['pdfarranger'],
    data_files=data_files,
    zip_safe=False,
    cmdclass={
        "build": build,
        "build_i18n": build_i18n,
        "build_icons": build_icons,
    },
    entry_points={
        'console_scripts': ['pdfarranger=pdfarranger.pdfarranger:main']
    },
    install_requires=['pikepdf>=6','python-dateutil>=2.4.0', 'packaging'],
    extras_require={
        'image': ['img2pdf>=0.3.4'],
    },
)
